#include "actions.hpp"
#include "ai.hpp"
#include "log.hpp"
#include "menu.hpp"
#include "pathfind.hpp"

namespace {

bool recruit(const gamemap& map, const gamemap::location& leader,
             const std::string& usage, const game_data& gameinfo,
			 int team_num, team& tm, int min_gold,
			 std::map<gamemap::location,unit>& units)
{
	log_scope("recruiting troops");
	
	//find an available unit that can be recruited, matches the desired
	//usage type, and comes in under budget
	const std::set<std::string>& recruits = tm.recruits();
	for(std::map<std::string,unit_type>::const_iterator i =
	    gameinfo.unit_types.begin(); i != gameinfo.unit_types.end(); ++i) {
		if(i->second.usage() == usage && recruits.count(i->second.name())
		   && tm.gold() - i->second.cost() > min_gold){
			
			tm.spend_gold(i->second.cost());

			return recruit_unit(map,team_num,units,i->second).empty();
		}
	}

	return false;
}

}

namespace ai {

void do_move(display& disp, const gamemap& map, const game_data& gameinfo,
             std::map<gamemap::location,unit>& units,
             std::vector<team>& teams, int team_num, const gamestatus& state)
{
	log_scope("doing ai move");

	team& current_team = teams[team_num-1];
	
	typedef gamemap::location location;
	typedef paths::route route;
		
	typedef std::map<location,paths> moves_map;
	moves_map possible_moves;
	for(std::map<gamemap::location,unit>::const_iterator un_it = units.begin();	
	    un_it != units.end(); ++un_it) {
		if(un_it->second.side() != team_num || un_it->second.can_recruit())
			continue;
		possible_moves.insert(std::pair<gamemap::location,paths>(
		                               un_it->first,paths(map,gameinfo,units,
		                               un_it->first)));
	}
	
	std::multimap<location,location> srcdst;
	std::multimap<location,location> dstsrc;
	
	for(moves_map::iterator m = possible_moves.begin();
	    m != possible_moves.end(); ++m) {
		for(std::map<location,route>::iterator rtit =
		    m->second.routes.begin(); rtit != m->second.routes.end();
			++rtit) {
			const location& src = m->first;
			const location& dst = rtit->first;

			if(src != dst && units.find(dst) == units.end()) {
				srcdst.insert(std::pair<location,location>(src,dst));
				dstsrc.insert(std::pair<location,location>(dst,src));
			}
		}
	}
	
	//no moves left, recruitment phase
	if(srcdst.empty()) {
		//take stock of our current set of units
		location leader;
		int num_units = 0;
		std::map<std::string,int> unit_types;
		for(std::map<location,unit>::const_iterator i = units.begin();
		    i != units.end(); ++i) {
			if(i->second.side() != team_num)
				continue;

			if(i->second.can_recruit()) {
				leader = i->first;
				continue;
			}
			
			unit_types[i->second.type().usage()]++;
			++num_units;
		}

		const int cash_flow = current_team.towers() - num_units;

		const int min_gold = 10 + (cash_flow < 0 ? -cash_flow*10 : 0);

		//count the number of scouts we have currently
		
		const int towers = map.towers().size();
		int taken_towers = 0;
		for(int j = 0; j != teams.size(); ++j) {
			taken_towers += teams[j].towers();
		}

		const int neutral_towers = towers - taken_towers;
		
		//we want at least one scout for every eight neutral towers
		int scouts_wanted = neutral_towers/8;
		if(scouts_wanted < 1)
			scouts_wanted = 1;
		while(unit_types["scout"] < scouts_wanted) {
			if(recruit(map,leader,"scout",gameinfo,team_num,current_team,
			           min_gold,units) == false)
				return;

			++unit_types["scout"];
		}

		//buy fighters as long as we have room and can afford it
		while(recruit(map,leader,"fighter",gameinfo,team_num,current_team,
		              min_gold,units)) {
		}
		
		return;
	}

	double best_ratio = 0.0;
	std::pair<location,location> best_move;
	location best_target;
	
	//look over possible targets to attack
	for(std::map<location,unit>::iterator u = units.begin();
	    u != units.end(); ++u) {

		//don't attack friendly units
		if(u->second.side() == team_num)
			continue;
			
		location adjacent[6];
		std::set<location> sources[6];
		int ntiles = 0;
		int max_units = 0;
		
		std::pair<location,location> attacker;
		std::string attacker_usage;
		
		get_adjacent_tiles(u->first,adjacent);
		std::set<location> potential_attackers;
		for(int i = 0; i != 6; ++i) {
			std::multimap<location,location>::iterator i1 =
			                               dstsrc.lower_bound(adjacent[i]);
			const std::multimap<location,location>::iterator i2 = 
			                               dstsrc.upper_bound(adjacent[i]);
			while(i1 != i2) {
				const std::string& new_usage
				       = units.find(i1->second)->second.type().usage();
				if(new_usage == "fighter" || attacker_usage != "fighter") {
					attacker_usage = new_usage;
					attacker.first = i1->second;
					attacker.second = i1->first;
				}
				
				sources[i].insert(i1->second);
				++i1;
			}

			if(sources[i].empty() == false)
				++ntiles;

			if(sources[i].size() > max_units)
				max_units = sources[i].size();
		}

		const double hitpoints = static_cast<double>(u->second.hitpoints());
		const double num_units =
		      static_cast<double>(max_units < ntiles ? max_units : ntiles);
		const double ratio = num_units/hitpoints;
		if(best_ratio > ratio)
			continue;

		if(attacker_usage != "fighter" && ratio < 0.2)
			continue;

		best_move = attacker;
		best_ratio = ratio;
		best_target = u->first;
	}

	//if we have found someone to attack 
	if(best_target.valid()) {
		
		paths& p = possible_moves.find(best_move.first)->second;

		std::map<location,route>::iterator rt = p.routes.begin();
		for(; rt != p.routes.end(); ++rt) {
			if(rt->first == best_move.second)
				break;
		}

		if(rt == p.routes.end()) {
			assert(false);
			return;
		}

		std::map<location,unit>::iterator u_it = units.find(best_move.first);
		unit current_unit = u_it->second;
		disp.scroll_to_tile(u_it->first.x,u_it->first.y);
		units.erase(u_it);
		
		std::vector<location> steps = rt->second.steps;
		steps.push_back(best_move.second); //add the destination to the steps
		disp.move_unit(steps,current_unit);
		current_unit.set_movement(0);
		units.insert(std::pair<location,unit>(best_move.second,current_unit));
		if(map[best_move.second.x][best_move.second.y] == gamemap::TOWER)
			get_tower(best_move.second,teams,team_num-1);

		disp.draw();

		const std::string& res = attack(map,best_move.second,best_target,
		                                0,units,state,gameinfo);
		show_message(300,300,disp.video(),res);
		
		do_move(disp,map,gameinfo,units,teams,team_num,state);
		return;
	}
	
	for(std::multimap<location,location>::const_iterator i = dstsrc.begin();
	    i != dstsrc.end(); ++i) {
		if(map[i->first.x][i->first.y] == gamemap::TOWER &&
		   !current_team.owns_tower(i->first)) {
			const std::map<location,unit>::iterator un = units.find(i->second);
			if(un == units.end()) {
				assert(false);
				return;
			}

			unit current_unit = un->second;
			disp.scroll_to_tile(un->first.x,un->first.y);
			units.erase(un);

			paths& p = possible_moves.find(i->second)->second;

			std::map<location,route>::iterator rt = p.routes.begin();
			for(; rt != p.routes.end(); ++rt) {
				if(rt->first == i->first)
					break;
			}


			if(rt == p.routes.end()) {
				assert(false);
				return;
			}

			std::vector<location> steps = rt->second.steps;
			steps.push_back(i->first); //add the destination to the steps
			disp.move_unit(steps,current_unit);
			current_unit.set_movement(0);
			units.insert(std::pair<location,unit>(i->first,current_unit));
			get_tower(i->first,teams,team_num-1);
			disp.draw();

			do_move(disp,map,gameinfo,units,teams,team_num,state);
			return;
		}
	}

	//just advance toward the enemy leader
	//
	//first find who the enemy leader is
	std::map<location,unit>::iterator u_it = units.begin();
	for(; u_it != units.end(); ++u_it) {
		if(u_it->second.can_recruit() && u_it->second.side() != team_num)
			break;
	}

	if(u_it == units.end())
		return;

	int closest_distance = 100000;
	std::pair<location,location> nearest;

	for(std::multimap<location,location>::const_iterator itor = dstsrc.begin();
	    itor != dstsrc.end(); ++itor) {
		const int distance = abs(itor->first.x - u_it->first.x) +
		                     abs(itor->first.y - u_it->first.y);
		if(distance < closest_distance) {
			nearest = *itor;
			closest_distance = distance;
		}
	}

	{
		const std::map<location,unit>::iterator unit_it =
		                                  units.find(nearest.second);
		if(unit_it == units.end()) {
			assert(false);
			return;
		}
		
		unit current_unit = unit_it->second;
		disp.scroll_to_tile(unit_it->first.x,unit_it->first.y);
		units.erase(unit_it);

		paths& p = possible_moves.find(nearest.second)->second;

		std::map<location,route>::iterator rt = p.routes.begin();
		for(; rt != p.routes.end(); ++rt) {
			if(rt->first == nearest.first)
				break;
		}


		if(rt == p.routes.end()) {
			assert(false);
			return;
		}

		std::vector<location> steps = rt->second.steps;
		steps.push_back(nearest.first); //add the destination to the steps
		disp.move_unit(steps,current_unit);
		current_unit.set_movement(0);
		units.insert(std::pair<location,unit>(nearest.first,current_unit));
		if(map[nearest.first.x][nearest.first.y] == gamemap::TOWER)
			get_tower(nearest.first,teams,team_num-1);
		
		disp.draw();

		do_move(disp,map,gameinfo,units,teams,team_num,state);
		return;
	}
}

}
