#include "SDL.h"

#include "actions.hpp"
#include "ai.hpp"
#include "config.hpp"
#include "display.hpp"
#include "key.hpp"
#include "menu.hpp"
#include "pathfind.hpp"
#include "team.hpp"
#include "unit_types.hpp"
#include "unit.hpp"
#include "video.hpp"

#include <fstream>
#include <iostream>
#include <sstream>

void play_level(game_data& gameinfo, config* level, CVideo& video)
{
	const std::string menu_items[] = {"Recruit","End Turn"};
	std::vector<std::string> menu;
	for(const std::string* menu_items_ptr = menu_items;
	    menu_items_ptr != menu_items + sizeof(menu_items)/sizeof(*menu_items);
		++menu_items_ptr)
		menu.push_back(*menu_items_ptr);
	
	const int num_turns = atoi(level->values["turns"].c_str());
	gamestatus status(num_turns);
	
	const gamemap map(read_file("data/maps/" + level->values["map"]));
	CKey key;
	typedef std::map<gamemap::location,unit> units_map;
	units_map units;

	std::vector<team> teams;
	
	std::vector<config*>& unit_cfg = level->children["side"];
	for(std::vector<config*>::iterator ui = unit_cfg.begin();
				ui != unit_cfg.end(); ++ui) {
		const unit new_unit(gameinfo, **ui);
		const std::string race_name = (*ui)->values["race"];
		const std::map<std::string,team::team_info>::const_iterator team_it =
				 gameinfo.races.find(race_name);
		if(team_it == gameinfo.races.end()) {
			std::cerr << "race not found: '" << race_name << "'\n";
			assert(false);
		}
		
		const team::team_info& info = team_it->second;
		
		units.insert(std::pair<gamemap::location,unit>(
					     map.starting_position(new_unit.side()), new_unit));
		teams.push_back(team(info));
	}
	
	display gui(units,video,map,status,teams);

	const double scroll_speed = 10.0;
	const double zoom_amount = 5.0;
	
	for(units_map::iterator i = units.begin(); i != units.end(); ++i) {
		i->second.new_turn();
	}
	
	bool left_button = false, right_button = false;
	
	paths current_paths;
	gamemap::location selected_hex;

	gui.scroll_to_tile(map.starting_position(1).x,map.starting_position(1).y,
	                   display::WARP);
	
	for(;;) {
		if(key[KEY_ESCAPE])
			break;

		int mousex, mousey;
		const int mouse_flags = SDL_GetMouseState(&mousex,&mousey);
		const bool new_left_button = mouse_flags & SDL_BUTTON_LMASK;
		const bool new_right_button = mouse_flags & SDL_BUTTON_RMASK;
		
		//highlight the hex that is currently moused over
		gui.highlight_hex(gui.hex_clicked_on(mousex,mousey));
			
		if(!left_button && new_left_button) {
			const gamemap::location& hex = gui.hex_clicked_on(mousex,mousey);

			const units_map::iterator u = units.find(selected_hex);

			//if we can move to that tile
			const std::map<gamemap::location,paths::route>::const_iterator
					route = current_paths.routes.find(hex);
			const units_map::iterator enemy = units.find(hex);

			//see if we're trying to attack an enemy
			if(route != current_paths.routes.end() && enemy != units.end() &&
			   hex != selected_hex &&
			   enemy->second.side() != u->second.side()) {
				const unit_type& type = u->second.type();
				const std::vector<attack_type>& attacks = type.attacks();
				std::vector<std::string> items;

				const std::map<std::string,unit_movement_type>::const_iterator
				   target_move_type = gameinfo.movement_types.find(
				                        enemy->second.type().movement_type());
				if(target_move_type == gameinfo.movement_types.end()) {
					assert(false);
				}

				for(std::vector<attack_type>::const_iterator a =
				    attacks.begin(); a != attacks.end(); ++a) {
					const double modifier = status.combat_modifier(
									                       type.alignment());

					int damage = (int)
						((double)target_move_type->second.damage_against(*a) *
							              modifier);
					if(damage < 1)
						damage = 1;

					std::stringstream att;
					att << a->name() << " (" << a->type() << ") "
					    << damage << "-" << a->num_attacks()
						<< (a->range() == attack_type::SHORT_RANGE ?
						    " Melee" : " Ranged");
					items.push_back(att.str());
				}

				items.push_back("Cancel");

				const int res = show_menu(300,300,video,items);

				if(res >= 0 && res < attacks.size()) {
					const std::string& msg = attack(map,selected_hex,hex,
					                                res,units,status,gameinfo);
					
					gui.draw(); //clear the screen
					show_message(300,300,video,msg);
					current_paths = paths();
					selected_hex = gamemap::location();
					gui.set_paths(NULL);
				}
			}
			
			//otherwise we're just trying to move to a hex
			else if(selected_hex.valid() && selected_hex != hex &&
				     enemy == units.end() &&
			         route != current_paths.routes.end()) {
				std::vector<gamemap::location> steps = route->second.steps;
				steps.push_back(hex);
				if(u != units.end()) {
					unit un = u->second;
					units.erase(u);
					gui.move_unit(steps,un);
					un.set_movement(route->second.move_left);

					//see if the unit has gotten a tower
					if(map[hex.x][hex.y] == gamemap::TOWER) {
						get_tower(hex,teams,0);
						un.set_movement(0); //end of turn if you get a tower
					}

					units.insert(std::pair<gamemap::location,unit>(hex,un));
				} else {
					assert(false);
				}


				selected_hex = gamemap::location();
				gui.select_hex(gamemap::location());
				gui.set_paths(NULL);
				current_paths = paths();

			} else {
				gui.set_paths(NULL);
				current_paths = paths();
			
				selected_hex = hex;
				gui.select_hex(hex);

				const units_map::iterator it = units.find(hex);
				if(it != units.end() && it->second.side() == 1) {
					current_paths = paths(map,gameinfo,units,hex);
					gui.set_paths(&current_paths);
				}
			}
		}
		
		left_button = new_left_button;

		if(!right_button && new_right_button) {
			if(!current_paths.routes.empty()) {
				selected_hex = gamemap::location();
				gui.select_hex(gamemap::location());
				gui.set_paths(NULL);
				current_paths = paths();
			} else {
				const int res = show_menu(300,300,video,menu);
				//end turn
				if(res == 1) {
					int ai_number = 2;
					for(std::vector<team>::iterator team_it = teams.begin()+1;
					    team_it != teams.end(); ++team_it, ++ai_number) {
						gui.scroll_to_tile(
						            map.starting_position(ai_number).x,
						            map.starting_position(ai_number).y);
						ai::do_move(gui,map,gameinfo,units,teams,
						            ai_number,status);
					}

					gui.scroll_to_tile(
					                map.starting_position(1).x,
					                map.starting_position(1).y);
					status.next_turn();
					std::map<int,int> expenditure;
					for(units_map::iterator i = units.begin();
					    i != units.end(); ++i) {
						i->second.new_turn();
						expenditure[i->second.side()]++;
					}

					int team_num = 1;
					for(std::vector<team>::iterator it = teams.begin();
					    it != teams.end(); ++it, ++team_num) {
						it->new_turn();
						it->spend_gold(expenditure[team_num]);
					}
				}

				//recruit
				else if(res == 0) {
						
					gui.draw(); //clear the old menu
					std::vector<std::string> item_keys;
					std::vector<std::string> items;
					const std::set<std::string>& recruits = teams[0].recruits();
					for(std::set<std::string>::const_iterator it =
					    recruits.begin(); it != recruits.end(); ++it) {
						item_keys.push_back(*it);
						const std::map<std::string,unit_type>::const_iterator
								u_type = gameinfo.unit_types.find(*it);
						if(u_type == gameinfo.unit_types.end()) {
							std::cerr << "could not find " << *it << std::endl;
							assert(false);
							continue;
						}
						
						const unit_type& type = u_type->second;
						std::stringstream description;
						description << *it << ": " << type.cost() << " gold";
						items.push_back(description.str());
					}

					const int recruit_res = show_menu(300,300,video,items);
					if(recruit_res != -1) {
						const std::string& name = item_keys[recruit_res];
						const std::map<std::string,unit_type>::const_iterator
								u_type = gameinfo.unit_types.find(name);
						assert(u_type != gameinfo.unit_types.end());
						
						if(u_type->second.cost() > teams[0].gold()) {
							show_message(300,300,video,
									 "Not enough gold!\n(Click to continue)");
						} else {
							const std::string& msg =
									recruit_unit(map,1,units,u_type->second);
							if(msg.empty())
								teams[0].spend_gold(u_type->second.cost());
							else
								show_message(300,300,video,msg);
						}
					}
				}
			}
		}

		right_button = new_right_button;
		
		if(key[KEY_UP] || mousey == 0)
			gui.scroll(0.0,-scroll_speed);

		if(key[KEY_DOWN] || mousey == gui.y()-1)
			gui.scroll(0.0,scroll_speed);

		if(key[KEY_LEFT] || mousex == 0)
			gui.scroll(-scroll_speed,0.0);

		if(key[KEY_RIGHT] || mousex == gui.x()-1)
			gui.scroll(scroll_speed,0.0);

		if(key[KEY_Z])
			gui.zoom(zoom_amount);

		if(key[KEY_X])
			gui.zoom(-zoom_amount);
		
		gui.draw();
	}
}

int main(int argc, char** argv)
{
	config game_config(read_file("data/game.cfg"));
	const std::vector<config*>& units = game_config.children["units"];
	if(units.empty()) {
		std::cerr << "Could not find units configuration\n";
		std::cerr << game_config.write();
		return 0;
	}

	const std::vector<config*>& scenarios = game_config.children["scenario"];

	game_data units_data(*units[0]);
	
	std::string text_chr = read_file("data/text.chr");
	text_chr.resize(256*8);
	
	CKey key;
	CVideo video(1024,768,16,FULL_SCREEN,text_chr.c_str());
	
	for(int i = 0; i != scenarios.size(); ++i) {
		play_level(units_data,scenarios[i],video);
	}

	return 0;
}
