#include "unit_types.hpp"

#include <cassert>
#include <cstdlib>

attack_type::attack_type(config& cfg) : cfg_(&cfg)
{}

const std::string& attack_type::name() const
{
	return cfg_->values["name"];
}

const std::string& attack_type::type() const
{
	return cfg_->values["type"];
}

attack_type::RANGE attack_type::range() const
{
	return cfg_->values["range"] == "long" ? LONG_RANGE : SHORT_RANGE;
}

int attack_type::damage() const
{
	return atol(cfg_->values["damage"].c_str());
}

int attack_type::num_attacks() const
{
	return atol(cfg_->values["number"].c_str());
}

unit_movement_type::unit_movement_type(config& cfg) : cfg_(cfg)
{}

const std::string& unit_movement_type::name() const
{
	return cfg_.values["name"];
}

int unit_movement_type::movement_cost(gamemap::TERRAIN terrain) const
{
	const std::vector<config*>& v = cfg_.children["movement costs"];
	if(v.empty())
		return 1;

	config* movement_costs = v[0];
	const std::string& name = gamemap::terrain_name(terrain);
	const std::string& val = movement_costs->values[name];
	return atol(val.c_str());
}

double unit_movement_type::defense_modifier(gamemap::TERRAIN terrain) const
{
	const std::vector<config*>& v = cfg_.children["defense"];
	if(v.empty())
		return 1;

	config* defense = v[0];
	const std::string& name = gamemap::terrain_name(terrain);
	const std::string& val = defense->values[name];
	return atof(val.c_str());
}

int unit_movement_type::damage_against(const attack_type& attack) const
{
	const std::vector<config*>& v = cfg_.children["resistance"];
	if(v.empty())
		return 1;

	config* resistance = v[0];
	const std::string& val = resistance->values[attack.type()];
	const double resist = atof(val.c_str());
	return static_cast<int>(resist * static_cast<double>(attack.damage()));
}

unit_type::unit_type(config& cfg) : cfg_(cfg)
{}

const std::string& unit_type::name() const
{
	return cfg_.values["name"];
}

const std::string& unit_type::image() const
{
	return cfg_.values["image"];
}

const std::string& unit_type::image_profile() const
{
	const std::string& val = cfg_.values["profile"];
	if(val.size() == 0)
		return image();
	else
		return val;
}

int unit_type::hitpoints() const
{
	return atoi(cfg_.values["hitpoints"].c_str());
}

std::vector<attack_type> unit_type::attacks() const
{
	std::vector<attack_type> res;
	const std::vector<config*>& v = cfg_.children["attack"];
	for(std::vector<config*>::const_iterator i = v.begin(); i != v.end(); ++i)
		res.push_back(attack_type(**i));

	return res;
}

const std::string& unit_type::movement_type() const
{
	return cfg_.values["movement_type"];
}

int unit_type::cost() const
{
	return atoi(cfg_.values["cost"].c_str());
}

int unit_type::experience_needed() const
{
	return atoi(cfg_.values["experience"].c_str());
}

const std::string& unit_type::advances_to() const
{
	return cfg_.values["advanceto"];
}

const std::string& unit_type::usage() const
{
	return cfg_.values["usage"];
}

int unit_type::level() const
{
	return atoi(cfg_.values["level"].c_str());
}

int unit_type::movement() const
{
	return atoi(cfg_.values["movement"].c_str());
}

unit_type::ALIGNMENT unit_type::alignment() const
{
	const std::string& align = cfg_.values["alignment"];
	if(align == "lawful")
		return LAWFUL;
	else if(align == "chaotic")
		return CHAOTIC;
	else
		return NEUTRAL;
}

const std::string& unit_type::alignment_description(unit_type::ALIGNMENT align)
{
	static const std::string aligns[] = { "Lawful", "Neutral", "Chaotic" };
	return aligns[align];
}

game_data::game_data(config& cfg) : cfg_(cfg)
{
	std::vector<config*>& move_types = cfg_.children["movetype"];
	for(std::vector<config*>::iterator i = move_types.begin();
					i != move_types.end(); ++i) {
		const unit_movement_type move_type(**i);
		movement_types.insert(
				std::pair<std::string,unit_movement_type>(move_type.name(),
						                                  move_type));
	}

	std::vector<config*>& u_types = cfg_.children["unit"];
	for(std::vector<config*>::iterator j = u_types.begin();
					j != u_types.end(); ++j) {
		const unit_type u_type(**j);
		unit_types.insert(
				std::pair<std::string,unit_type>(u_type.name(),u_type));
	}

	std::vector<config*>& race_types = cfg_.children["race"];
	for(std::vector<config*>::iterator k = race_types.begin();
					k != race_types.end(); ++k) {
		const team::team_info race(**k);
		races.insert(
				std::pair<std::string,team::team_info>(race.name,race));
	}
}
