#include "actions.hpp"
#include "pathfind.hpp"

#include <string>
#include <sstream>

std::string recruit_unit(const gamemap& map, int side,
                         std::map<gamemap::location,unit>& units, unit& new_unit)
{
	typedef std::map<gamemap::location,unit> units_map;

	gamemap::location recruit_location;

	//find the unit that can recruit
	for(units_map::const_iterator u = units.begin(); u != units.end(); ++u) {
		if(u->second.can_recruit() && u->second.side() == side) {
			gamemap::location adjacent[6];

			//see if there is an adjacent tile that the
			//unit can be placed onto
			get_adjacent_tiles(u->first,adjacent);
			for(int i = 0; i != 6; ++i) {
				if(map.on_board(adjacent[i]) &&
				   map[adjacent[i].x][adjacent[i].y] == gamemap::CASTLE &&
				   units.find(adjacent[i]) == units.end()) {
					recruit_location = adjacent[i];
					break;
				}
			}

			break;
		}
	}

	if(!recruit_location.valid()) {
		return "To recruit a unit, you must have a vacant castle\n"
				"tile adjacent to your leader\n";
	}

	new_unit.set_movement(0);
	new_unit.set_attacked();
	units.insert(std::pair<gamemap::location,unit>(
							recruit_location,new_unit));

	return std::string();
}

std::string attack(const gamemap& map,
                   const gamemap::location& attacker,
                   const gamemap::location& defender,
				   int attack_with,
				   std::map<gamemap::location,unit>& units,
				   const gamestatus& state,
				   const game_data& info, bool player_is_attacker)
{
	const std::map<gamemap::location,unit>::iterator a = units.find(attacker);
	const std::map<gamemap::location,unit>::iterator d = units.find(defender);

	int attackerxp = d->second.type().level()*2;
	int defenderxp = a->second.type().level()*2;

	a->second.set_attacked();
	
	const gamemap::TERRAIN attacker_terrain = map[attacker.x][attacker.y];
	const gamemap::TERRAIN defender_terrain = map[defender.x][defender.y];

	const std::map<std::string,unit_movement_type>::const_iterator
			attacker_move_type = info.movement_types.find(
							a->second.type().movement_type());

	const double chance_to_hit_attacker =
			attacker_move_type->second.defense_modifier(attacker_terrain);
	
	const std::map<std::string,unit_movement_type>::const_iterator
			defender_move_type = info.movement_types.find(
							d->second.type().movement_type());

	const double chance_to_hit_defender =
			defender_move_type->second.defense_modifier(defender_terrain);

	const std::vector<attack_type>& attacker_attacks =
			a->second.type().attacks();
	const std::vector<attack_type>& defender_attacks =
			d->second.type().attacks();

	const attack_type& attack = attacker_attacks[attack_with];
	int nattacks = attack.num_attacks();
	int defend;
	int ndefends = 0;
	for(defend = 0; defend != defender_attacks.size(); ++defend) {
		if(defender_attacks[defend].range() == attack.range())
			break;
	}

	const bool counterattack = defend != defender_attacks.size();
	
	int damage_attacker_takes = 0;
	if(counterattack) {
		damage_attacker_takes = int(double(
		   attacker_move_type->second.damage_against(defender_attacks[defend]))
		      * state.combat_modifier(d->second.type().alignment()));
		if(damage_attacker_takes < 1)
			damage_attacker_takes = 1;

		ndefends = defender_attacks[defend].num_attacks();
	}

	const int damage_defender_takes = int(
			double(defender_move_type->second.damage_against(attack))
			  * state.combat_modifier(a->second.type().alignment()));
	
	std::stringstream report;

	report << "attacker: " << damage_defender_takes << "-"
	       << attack.num_attacks() << " "
		   << int(100.0*chance_to_hit_defender) << "%\n"
		   << "defender: " << damage_attacker_takes << "-"
		   << (counterattack ? defender_attacks[defend].num_attacks() : 0)
		   << " "
		   << (counterattack ? int(100.0*chance_to_hit_attacker) : 0)
		   << "%\n";

	const char attacker_hits = player_is_attacker ? '@' : '#';
	const char defender_hits = player_is_attacker ? '#' : '@';
	
	while(nattacks > 0 || ndefends > 0) {
		if(nattacks > 0) {
			const double roll = double(rand()%1000)/1000.0;
			const bool hits = roll < chance_to_hit_defender;
			if(hits) {
				report << attacker_hits;
				report << "attacker hits for " << damage_defender_takes;
				const bool dies = d->second.gets_hit(damage_defender_takes);
				++attackerxp;
				if(dies) {
					attackerxp = 10*d->second.type().level();
					defenderxp = 0;
					report << ", defender is killed\n";
					units.erase(d);
					break;
				} else {
					report << ", defender has " << d->second.hitpoints()
					       << " hitpoints remaining\n";
				}
			} else {
				report << "attacker misses\n";
			}

			--nattacks;
		}

		if(ndefends > 0) {
			const double roll = double(rand()%1000)/1000.0;
			const bool hits = roll < chance_to_hit_attacker;

			if(hits) {
				report << defender_hits;
				report << "defender hits for " << damage_attacker_takes;
				const bool dies = a->second.gets_hit(damage_attacker_takes);
				++defenderxp;
				if(dies) {
					defenderxp = 10*a->second.type().level();
					attackerxp = 0;
					report << ", attacker is killed\n";
					units.erase(a);
					break;
				} else {
					report << ", attacker has " << a->second.hitpoints()
					       << " hitpoints remaining\n";
				}
			} else {
				report << "defender misses\n";
			}

			--ndefends;
		}
	}

	if(attackerxp) {
		report << "attacker is awarded " << attackerxp << " experience\n";
		const std::string advance_to = a->second.type().advances_to();
		if(a->second.get_experience(attackerxp) && advance_to != "null") {
			const std::map<std::string,unit_type>::const_iterator new_type =
			     info.unit_types.find(advance_to);
			if(new_type != info.unit_types.end()) {
				const gamemap::location loc = a->first;
				const int side = a->second.side();
				units.erase(a);
				units.insert(std::pair<gamemap::location,unit>(loc,
				                 unit(&(new_type->second),a->second.side())));
				report << "attacker advances to " << advance_to << "!\n";
			}
		}
	}

	if(defenderxp) {
		report << "defender is awarded " << defenderxp << " experience\n";
		const std::string advance_to = d->second.type().advances_to();
		if(d->second.get_experience(defenderxp) && advance_to != "null") {
			const std::map<std::string,unit_type>::const_iterator new_type =
			     info.unit_types.find(advance_to);
			if(new_type != info.unit_types.end()) {
				const gamemap::location loc = d->first;
				const int side = d->second.side();
				units.erase(d);
				units.insert(std::pair<gamemap::location,unit>(loc,
				                 unit(&(new_type->second),d->second.side())));
				report << "defender advances to " << advance_to << "!\n";
			}
		}
	}

	return report.str();
}

void get_tower(const gamemap::location& loc, std::vector<team>& teams,
               int team_num)
{
	for(int i = 0; i != teams.size(); ++i) {
		if(i != team_num && teams[i].owns_tower(loc)) {
			teams[i].lose_tower(loc);
		}
	}

	teams[team_num].get_tower(loc);
}

void calculate_healing(const gamemap& map,
                       std::map<gamemap::location,unit>& units, int side)
{
	for(std::map<gamemap::location,unit>::iterator i = units.begin();
	    i != units.end(); ++i) {
		if(i->second.side() != side)
			continue;

		bool heals = false;
		if(map[i->first.x][i->first.y] == gamemap::TOWER) {
			heals = true;
		}

		if(i->second.type().regenerates()) {
			heals = true;
		}

		gamemap::location adjacent[6];
		get_adjacent_tiles(i->first,adjacent);
		for(int j = 0; j != 6; ++j) {
			const std::map<gamemap::location,unit>::const_iterator adj_unit =
			                                           units.find(adjacent[j]);
			if(adj_unit != units.end() && adj_unit->second.type().heals() &&
			   adj_unit->second.side() == side) {
				heals = true;
			}
		}

		if(heals) {
			i->second.heal();
		}
	}
}
