#include "display.hpp"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iostream>
#include <sstream>

display::display(const unit_map& units, CVideo& video, const gamemap& map,
				 const gamestatus& status, const std::vector<team>& t)
		             : screen_(video), xpos_(0.0), ypos_(0.0), zoom_(100.0),
                       map_(map), units_(units), energy_bar_count_(-1),
                       minimap_(NULL), pathsList_(NULL), status_(status),
                       teams_(t), lastDraw_(0), drawSkips_(0)
{
}

void clear_surfaces(std::map<std::string,SDL_Surface*>& surfaces)
{
	for(std::map<std::string,SDL_Surface*>::iterator i = surfaces.begin();
					i != surfaces.end(); ++i) {
		SDL_FreeSurface(i->second);
	}

	surfaces.clear();
}

display::~display()
{
	clear_surfaces(images_);
	clear_surfaces(scaledImages_);
	clear_surfaces(greyedImages_);
	clear_surfaces(brightenedImages_);
	SDL_FreeSurface(minimap_);
}

int display::x() const { return screen_.getx(); }
int display::mapx() const { return x() - 120; }
int display::y() const { return screen_.gety(); }

void display::select_hex(gamemap::location hex)
{
	selectedHex_ = hex;
}

void display::highlight_hex(gamemap::location hex)
{
	mouseoverHex_ = hex;
}

gamemap::location display::hex_clicked_on(int xclick, int yclick)
{
	if(xclick > mapx())
		return gamemap::location();
		
	const double xtile = xpos_/(zoom_*0.75) +
			               static_cast<double>(xclick)/(zoom_*0.75) - 0.25;
	const double ytile = ypos_/zoom_ + static_cast<double>(yclick)/zoom_
	                      + (((static_cast<int>(xtile)%2) == 1) ? -0.5:0.0);

	return gamemap::location(static_cast<int>(xtile),static_cast<int>(ytile));
}

void display::scroll(double xmove, double ymove)
{
	xpos_ += xmove;
	ypos_ += ymove;
	bounds_check_position();
}

void display::zoom(double amount)
{
	clear_surfaces(scaledImages_);
	clear_surfaces(greyedImages_);
	clear_surfaces(brightenedImages_);
	energy_bar_count_ = -1;

	xpos_ /= zoom_;
	ypos_ /= zoom_;
	
	const double max_zoom = 200.0;
	zoom_ += amount;
	if(zoom_ > max_zoom)
		zoom_ = max_zoom;

	xpos_ *= zoom_;
	ypos_ *= zoom_;

	bounds_check_position();
}

void display::scroll_to_tile(int x, int y, SCROLL_TYPE scroll_type)
{
	const double xpos = static_cast<double>(x)*zoom_*0.75 - xpos_;
	const double ypos = static_cast<double>(y)*zoom_ - ypos_ +
					                    ((x % 2) == 1 ? zoom_/2.0 : 0.0);

	const double speed = 25.0;

	const double desiredxpos = this->mapx()/2.0 - zoom_/2.0;
	const double desiredypos = this->   y()/2.0 - zoom_/2.0;

	const double xmove = xpos - desiredxpos;
	const double ymove = ypos - desiredypos;

	int num_moves = static_cast<int>((abs(xmove) > abs(ymove) ?
	                                  abs(xmove):abs(ymove))/speed);
	if(scroll_type == WARP)
		num_moves = 1;

	const double divisor = static_cast<double>(num_moves);
	
	for(int i = 0; i != num_moves; ++i) {
		scroll(xmove/divisor,ymove/divisor);
		draw();
	}
}

void display::bounds_check_position()
{
	const double min_zoom1 = static_cast<double>(mapx()/(map_.x()*0.75 + 0.25));
	const double min_zoom2 = static_cast<double>(y()/map_.y());
	const double min_zoom = min_zoom1 > min_zoom2 ? min_zoom1 : min_zoom2;
	const double max_zoom = 200.0;

	zoom_ = floor(zoom_);
	
	if(zoom_ < min_zoom)
		zoom_ = min_zoom;

	if(zoom_ > max_zoom)
		zoom_ = max_zoom;

	const double xend = zoom_*map_.x()*0.75 + zoom_*0.25;
	const double yend = zoom_*map_.y() + zoom_/2.0;

	if(xpos_ + static_cast<double>(mapx()) > xend)
		xpos_ -= xpos_ + static_cast<double>(mapx()) - xend;

	if(ypos_ + static_cast<double>(y()) > yend)
		ypos_ -= ypos_ + static_cast<double>(y()) - yend;
	
	if(xpos_ < 0.0)
		xpos_ = 0.0;

	if(ypos_ < 0.0)
		ypos_ = 0.0;
}

void display::draw(bool update,bool force)
{
	for(int x = -1; x <= map_.x(); ++x)
		for(int y = -1; y <= map_.y(); ++y)
			draw_tile(x,y);

	draw_sidebar();

	const int max_skips = 5;
	const int time_between_draws = 20;
	const int current_time = SDL_GetTicks();
	const int wait_time = lastDraw_ + time_between_draws - current_time;
	if(wait_time > 0) {
		SDL_Delay(wait_time);
	}
	
	if(update) {
		lastDraw_ = SDL_GetTicks();
		
		if(wait_time >= 0 || drawSkips_ >= max_skips || force)
			update_display();
		else
			drawSkips_++;

	}
}

void display::update_display()
{
	screen_.update(0,0,this->x(),this->y());
}

void display::draw_sidebar()
{
	draw_minimap(mapx()+10,10,100,200);
	draw_unit_details(mapx()+10,400);
	draw_game_status(mapx()+10,200);
}

void display::draw_game_status(int x, int y)
{
	std::stringstream details;
	details << gamestatus::timeofdayDescription(status_.timeofday()) << "  \n"
			<< "Turn: " << status_.turn() << "/"
			<< status_.number_of_turns() << "\nGold: " << teams_[0].gold()
			<< "  \n"
			<< "Villages: " << teams_[0].towers() << "  \n";

	if(map_.on_board(mouseoverHex_)) {
		const gamemap::TERRAIN terrain = map_[mouseoverHex_.x][mouseoverHex_.y];
		const char* const terrain_name = gamemap::terrain_name(terrain);
		std::string name(terrain_name);
		if(!name.empty())
			name[0] = toupper(name[0]);
		details << name << "    \n";
	} else {
		details << "                \n";
	}

	const int bgcolour = 0;
	const int colour = 0xffffff;
	screen_.drawText(x,y,colour,bgcolour,details.str().c_str());
}

void display::draw_unit_details(int x, int y)
{
	if(!selectedHex_.valid())
		return;

	const unit_map::const_iterator i = units_.find(selectedHex_);
	if(i == units_.end())
		return;

	const unit& u = i->second;

	const int bgcolour = 0;
	const int colour = 0xffffff;
	
	std::stringstream details;
	details << u.description() << "             \n"
	        << u.type().name() << "      \n"
			<< unit_type::alignment_description(u.type().alignment()) << "\n"
			<< "hp: " << u.hitpoints() << "/" << u.type().hitpoints()<<"\n"
			<< "xp: " << u.experience() << "/"
			<< u.type().experience_needed() << "\n"
			<< "moves: " << u.movement_left() << "/" <<u.type().movement()<<"\n";
	
	const std::vector<attack_type>& attacks = u.type().attacks();
	for(std::vector<attack_type>::const_iterator at_it = attacks.begin();
	    at_it != attacks.end(); ++at_it) {
		details << "\n" << at_it->name() << " (" << at_it->type() << ")    \n"
		        << (at_it->range() == attack_type::SHORT_RANGE ?
		            "melee" : "ranged") << "          \n"
				<< at_it->damage() << "-" << at_it->num_attacks()
		        << "      \n";
	}

	for(int j = 0; j != 2 - attacks.size(); ++j) {
		details << "\n             \n            \n               \n";
	}
	
	y += screen_.drawText(x,y,colour,bgcolour,details.str().c_str());

	SDL_Surface* const profile =
			getImage(u.type().image_profile(),UNSCALED);

	if(profile == NULL)
		return;

	SDL_Surface* const screen = screen_.getSurface();
	SDL_Rect dstrect;
	dstrect.x = x;
	dstrect.y = y;
	dstrect.w = profile->w;
	dstrect.h = profile->h;
	if(x + profile->w < screen->w && y + profile->h < screen->h)
		SDL_BlitSurface(profile,NULL,screen,&dstrect);
}

void display::draw_minimap(int x, int y, int w, int h)
{
	SDL_Surface* const surface = getMinimap(w,h);
	if(surface == 0)
		return;
	
	if(w > surface->w)
		w = surface->w;

	if(h > surface->h)
		h = surface->h;

	if(w < surface->w || h < surface->h)
		return;

	SDL_Rect dest;
	dest.x = x;
	dest.y = y;
	dest.w = w;
	dest.h = h;

	SDL_BlitSurface(surface,NULL,screen_.getSurface(),&dest);

	const double scaling = static_cast<double>(surface->w/map_.x());

	const int xbox = static_cast<int>(scaling*xpos_/(zoom_*0.75));
	const int ybox = static_cast<int>(scaling*ypos_/zoom_);
	
	const int wbox = static_cast<int>(scaling*mapx()/(zoom_*0.75) - scaling);
	const int hbox = static_cast<int>(scaling*this->y()/zoom_ - scaling);

	const short boxcolour = (short)0xFFFF;
	SDL_Surface* const screen = screen_.getSurface();
	short* const data = reinterpret_cast<short*>(screen->pixels);
	short* const start_top = data + (y+ybox)*screen->w + (x+xbox);
	short* const end_top = start_top + wbox;

	std::fill(start_top,end_top,boxcolour);

	short* const start_bot = start_top + hbox*screen->w;
	short* const end_bot = start_bot + wbox;

	std::fill(start_bot,end_bot,boxcolour);

	short* side;
	for(side = start_top; side != start_bot; side += screen->w) {
		*side = boxcolour;
		side[wbox] = boxcolour;
	}
}

void display::draw_tile(int x, int y)
{
	int xpos = static_cast<int>(static_cast<double>(x)*zoom_*0.75 - xpos_);
	int ypos = static_cast<int>(static_cast<double>(y)*zoom_ - ypos_ +
					                    ((x % 2) == 1 ? zoom_/2.0 : 0.0));
	
	if(xpos > mapx() || ypos > this->y())
		return;
	
	int xend = xpos + static_cast<int>(zoom_);
	int yend = ypos + static_cast<int>(zoom_);

	if(xend < 0 || yend < 0)
		return;

	gamemap::TERRAIN terrain = gamemap::VOID_TERRAIN;

	if(x >= 0 && y >= 0 && x < map_.x() && y < map_.y())
		terrain = map_[x][y];

	IMAGE_TYPE image_type = SCALED;
	
	//find if this tile should be greyed
	if(pathsList_ != NULL && pathsList_->routes.find(gamemap::location(x,y)) ==
					         pathsList_->routes.end()) {
		image_type = GREYED;
	} else if(gamemap::location(x,y) == mouseoverHex_) {
		image_type = BRIGHTENED;
	}
	
	SDL_Surface* surface = NULL;
	int r = 0, g = 0, b = 0;
	switch(terrain) {
		case gamemap::VOID_TERRAIN:
			surface = getImage("void.bmp",image_type);
			break;
		case gamemap::DEEP_WATER:
			surface = getImage("ocean.bmp",image_type);
			break;
		case gamemap::SHALLOW_WATER:
			surface = getImage("coast.bmp",image_type);
			break;
		case gamemap::GRASSLAND:
			surface = getImage("grassland.bmp",image_type);
			break;
		case gamemap::DESERT:
			surface = getImage("desert.bmp",image_type);
			break;
		case gamemap::FOREST:
			surface = getImage("forest.bmp",image_type);
			break;
		case gamemap::MOUNTAINS:
			surface = getImage("mountains.bmp",image_type);
			break;
		case gamemap::TOWER:
		{
			for(int i = 0; i != teams_.size(); ++i) {
				if(teams_[i].owns_tower(gamemap::location(x,y))) {
					std::stringstream image_name;
					image_name << "flag-team" << (i+1) << ".bmp";
					surface = getImage(image_name.str(),image_type);
					assert(surface);
					break;
				}
			}
			if(surface == NULL)
				surface = getImage("flag-neutral.bmp",image_type);
			break;
		}
		case gamemap::CASTLE:
			surface = getImage("castle.bmp",image_type);
			break;
	}

	assert(surface != NULL);
	if(surface == NULL)
		return;

	int ysrc = 0, xsrc = 0;
	if(ypos < 0) {
		ysrc -= ypos;
		ypos = 0;
	}

	if(xpos < 0) {
		xsrc -= xpos;
		xpos = 0;
	}

	if(xend >= mapx())
		xend = mapx()-1;

	if(yend >= this->y())
		yend = this->y()-1;

	if(xend < xpos || yend < ypos)
		return;

	SDL_Surface* unit_image = NULL;
	SDL_Surface* energy_image = NULL;

	bool selected = false;

	double unit_energy = 0.0;
	
	const short energy_loss_colour = 0;
	short energy_colour = 0;

	//see if there is a unit on this tile
	const unit_map::const_iterator it = units_.find(gamemap::location(x,y));
	if(it != units_.end()) {
		unit_image = getImage(it->second.type().image());
		const int unit_move = it->second.movement_left();
		const int unit_total_move = it->second.type().movement();
		energy_image = getImage(
		       unit_move == unit_total_move ? "unmoved-energy.bmp" :
		      (unit_move > 0 ? "partmoved-energy.bmp" : "moved-energy.bmp"));
		unit_energy = double(it->second.hitpoints()) /
		              double(it->second.type().hitpoints());
		selected = (gamemap::location(x,y) == selectedHex_);

		{
			int er = 0;
			int eg = 0;
			int eb = 0;
			if(unit_energy < 0.33) {
				er = 200;
			} else if(unit_energy < 0.66) {
				er = 200;
				eg = 200;
			} else {
				eg = 200;
			}

			energy_colour = ((er >> energy_image->format->Rloss) <<
			                 energy_image->format->Rshift) +
			                ((eg >> energy_image->format->Gloss) <<
							 energy_image->format->Gshift) +
			                ((eb >> energy_image->format->Bloss) <<
							 energy_image->format->Bshift);
		}
	}
	
	double total_energy = double(calculate_energy_bar());
	double energy_upto = 0.0;

	SDL_Surface* const dst = screen_.getSurface();
	for(int j = ypos; j != yend; ++j) {
		const int yloc = ysrc+j-ypos;
		const int xoffset = abs(yloc - static_cast<int>(zoom_/2.0))/2;
		int xdst = xpos;
		if(xoffset > xsrc) {
			xdst += xoffset - xsrc;
			if(xend < xdst)
				continue;
		}

		const int maxlen = static_cast<int>(zoom_) - xoffset*2;
		const int len = ((xend - xdst) > maxlen) ? maxlen : xend - xdst;
		
		short* startsrc = reinterpret_cast<short*>(surface->pixels) +
			                yloc*surface->w + (xoffset > xsrc ? xoffset:xsrc);
		short* endsrc = startsrc + len;
		short* startdst =
			reinterpret_cast<short*>(dst->pixels) + j*dst->w + xdst;
		std::copy(startsrc,endsrc,startdst);

		if(unit_image != NULL && energy_image != NULL) {
			const short replace_energy = 0xFFFF;
			
			//we want to count the total number of white pixels in the energy
			//image. White pixels get replaced with black for energy taken,
			//and green for energy remaining
			short* startenergy = reinterpret_cast<short*>(energy_image->pixels)+
			           yloc*energy_image->w + (xoffset > xsrc ? xoffset:xsrc);

			startsrc = reinterpret_cast<short*>(unit_image->pixels) +
					      yloc*unit_image->w + (xoffset > xsrc ? xoffset:xsrc);
			endsrc = startsrc + len;

			if(!selected) {
				while(startsrc != endsrc) {
					if(*startenergy != 0) {
						if(*startenergy == replace_energy) {
							if(1.0 - energy_upto/total_energy <= unit_energy) {
								*startdst = energy_colour;
							} else {
								*startdst = energy_loss_colour;
							}
							energy_upto += 1.0;
						} else {
							*startdst = *startenergy;
						}
					} else if(*startsrc != 0) {
						*startdst = *startsrc;
					}

					++startdst;	
					++startsrc;
					++startenergy;
				}
			} else {
				const int maxred   = unit_image->format->Rmask >>
						             unit_image->format->Rshift;
				const int maxgreen = unit_image->format->Gmask >>
						             unit_image->format->Gshift;
				const int maxblue  = unit_image->format->Bmask >>
						             unit_image->format->Bshift;

				while(startsrc != endsrc) {
					if(*startenergy != 0) {
						if(*startenergy == replace_energy) {
							if(1.0 - energy_upto/total_energy <= unit_energy) {
								*startdst = energy_colour;
							} else {
								*startdst = energy_loss_colour;
							}
							energy_upto += 1.0;
						} else {
							*startdst = *startenergy;
						}

						++startsrc;
						++startdst;
						++startenergy;
						continue;
					}

					if(*startsrc == 0) {
						++startsrc;
						++startdst;
						++startenergy;
						continue;
					}
						
					int r = (*startsrc&unit_image->format->Rmask) >>
							          unit_image->format->Rshift;
					int g = (*startsrc&unit_image->format->Gmask) >>
							          unit_image->format->Gshift;
					int b = (*startsrc&unit_image->format->Bmask) >>
							          unit_image->format->Bshift;

					r = static_cast<int>(static_cast<double>(r) * 1.5);
					if(r > maxred)
						r = maxred;
					g = static_cast<int>(static_cast<double>(g) * 1.5);
					if(g > maxgreen)
						g = maxgreen;
					b = static_cast<int>(static_cast<double>(b) * 1.5);
					if(b > maxblue)
						b = maxblue;

					const int col = (r << unit_image->format->Rshift) |
							        (g << unit_image->format->Gshift) |
									(b << unit_image->format->Bshift);

					*startdst = col;
					++startdst;
					++startsrc;
					++startenergy;
				}
			}
		}
	}
}

SDL_Surface* scale_surface(SDL_Surface* surface, int w, int h)
{
	SDL_Surface* const dest = SDL_CreateRGBSurface(SDL_SWSURFACE,w,h,
					                     surface->format->BitsPerPixel,
										 surface->format->Rmask,
										 surface->format->Gmask,
										 surface->format->Bmask,
										 surface->format->Amask);
	if(dest == NULL)
		return NULL;

	const double xratio = static_cast<double>(surface->w)/
			              static_cast<double>(w);
	const double yratio = static_cast<double>(surface->h)/
			              static_cast<double>(h);
	
	double ysrc = 0.0;
	for(int ydst = 0; ydst != h; ++ydst, ysrc += yratio) {
		double xsrc = 0.0;
		for(int xdst = 0; xdst != w; ++xdst, xsrc += xratio) {
			const int xsrcint = static_cast<int>(xsrc);
			const int ysrcint = static_cast<int>(ysrc);

			reinterpret_cast<short*>(dest->pixels)[ydst*w + xdst] =
			reinterpret_cast<short*>(surface->pixels)[
			                           ysrcint*surface->w + xsrcint];
		}
	}

	return dest;
}

void display::blit_surface(int x, int y, SDL_Surface* surface)
{
	SDL_Surface* const target = video().getSurface();

	const int srcx = x < 0 ? -x : 0;
	const int srcw = x + surface->w > target->w ? target->w - x :
	                                              surface->w - srcx;
	const int srcy = y < 0 ? -y : 0;
	const int srch = y + surface->h > target->h ? target->h - x :
	                                              surface->h - srcy;

	if(srcw <= 0 || srch <= 0 || srcx >= surface->w || srcy >= surface->h)
		return;

	SDL_LockSurface(surface);
	SDL_LockSurface(target);
	
	if(x < 0)
		x = 0;

	if(y < 0)
		y = 0;
	
	//lines are padded to always fit on 4-byte boundaries, so see if there
	//is padding at the beginning of every line
	const int padding = (surface->w%2) == 1 ? 1:0;
	const int surface_width = surface->w + padding;
	
	const short* src = reinterpret_cast<short*>(surface->pixels) +
	                   srcy*surface_width + srcx + padding;
	short* dst = reinterpret_cast<short*>(target->pixels) + y*target->w + x;

	static const short transperant = 0;
	
	for(int i = 0; i != srch; ++i) {
		const short* s = src + i*surface_width + padding;
		const short* const end = s + srcw;
		short* d = dst + i*target->w;
		while(s != end) {
			if(*s != transperant) {
				*d = *s;
			}
			
			++s;
			++d;
		}
	}
	SDL_UnlockSurface(surface);
	SDL_UnlockSurface(target);
}

SDL_Surface* display::getImage(const std::string& filename,
				               display::IMAGE_TYPE type)
{
	if(type == GREYED)
		return getImageTinted(filename,GREY_IMAGE);
	else if(type == BRIGHTENED)
		return getImageTinted(filename,BRIGHTEN_IMAGE);
		
	std::map<std::string,SDL_Surface*>::iterator i;
	
	if(type == SCALED) {
		i = scaledImages_.find(filename);
		if(i != scaledImages_.end())
			return i->second;
	}

	i = images_.find(filename);

	if(i == images_.end()) {
		const std::string images_path = "images/";
		const std::string images_filename = images_path + filename;
		SDL_Surface* const surf = SDL_LoadBMP(images_filename.c_str());
		if(surf == NULL) {
			assert(false);
			return NULL;
		}

		SDL_Surface* const conv = SDL_ConvertSurface(surf,
						                     screen_.getSurface()->format,
											 SDL_SWSURFACE);
		SDL_FreeSurface(surf);

		i = images_.insert(std::pair<std::string,SDL_Surface*>(filename,conv))
				               .first;
	}

	if(type == UNSCALED) {
		return i->second;
	}

	const int z = static_cast<int>(zoom_);
	SDL_Surface* const new_surface = scale_surface(i->second,z,z);

	if(new_surface == NULL)
		return NULL;
	
	scaledImages_.insert(std::pair<std::string,SDL_Surface*>(filename,
								                             new_surface));
	return new_surface;
}

SDL_Surface* display::getImageTinted(const std::string& filename, TINT tint)
{
	std::map<std::string,SDL_Surface*>& image_map =
	       tint == GREY_IMAGE ? greyedImages_ : brightenedImages_;
	
	const std::map<std::string,SDL_Surface*>::iterator itor =
			image_map.find(filename);

	if(itor != image_map.end())
		return itor->second;
		
	SDL_Surface* const base = getImage(filename,SCALED);
	if(base == NULL)
		return NULL;

	SDL_Surface* const surface = 
		           SDL_CreateRGBSurface(SDL_SWSURFACE,base->w,base->h,
					                     base->format->BitsPerPixel,
										 base->format->Rmask,
										 base->format->Gmask,
										 base->format->Bmask,
										 base->format->Amask);
	
	image_map.insert(std::pair<std::string,SDL_Surface*>(filename,surface));
	
	short* begin = reinterpret_cast<short*>(base->pixels);
	const short* const end = begin + base->w*base->h;
	short* dest = reinterpret_cast<short*>(surface->pixels);

	const int rmax = 0xFF >> base->format->Rloss;
	const int gmax = 0xFF >> base->format->Gloss;
	const int bmax = 0xFF >> base->format->Bloss;
	
	while(begin != end) {
		int r = (*begin & base->format->Rmask) >> base->format->Rshift;
		int g = (*begin & base->format->Gmask) >> base->format->Gshift;
		int b = (*begin & base->format->Bmask) >> base->format->Bshift;
		
		if(tint == GREY_IMAGE) {
			const double greyscale = (double(r)/(double)rmax +
					                  double(g)/(double)gmax +
							          double(b)/(double)bmax)/3.0;
	
			r = (int)(rmax*greyscale);
			g = (int)(gmax*greyscale);
			b = (int)(bmax*greyscale);
		} else {
			r = int(double(r)*1.5);
			g = int(double(g)*1.5);
			b = int(double(b)*1.5);

			if(r > rmax)
				r = rmax;

			if(g > gmax)
				g = gmax;

			if(b > bmax)
				b = bmax;
		}

		*dest = (r << base->format->Rshift) |
				(g << base->format->Gshift) |
				(b << base->format->Bshift);
		++dest;
		++begin;
	}

	return surface;
}

SDL_Surface* display::getMinimap(int w, int h)
{
	if(minimap_ == NULL) {
		SDL_Surface* const surface = screen_.getSurface();
		minimap_ = SDL_CreateRGBSurface(SDL_SWSURFACE,map_.x(),map_.y(),
					                     surface->format->BitsPerPixel,
										 surface->format->Rmask,
										 surface->format->Gmask,
										 surface->format->Bmask,
										 surface->format->Amask);
		if(minimap_ == NULL)
			return NULL;

		short* data = reinterpret_cast<short*>(minimap_->pixels);
		for(int y = 0; y != map_.y(); ++y) {
			for(int x = 0; x != map_.x(); ++x) {
				int r = 0;
				int g = 0;
				int b = 0;
				const gamemap::TERRAIN terrain = map_[x][y];
				switch(terrain) {
					case gamemap::DEEP_WATER:
						b = 100;
						break;
					case gamemap::SHALLOW_WATER:
						b = 200;
						break;
					case gamemap::GRASSLAND:
						g = 150;
						break;
					case gamemap::DESERT:
						r = 200;
						g = 200;
						break;
					case gamemap::FOREST:
						g = 100;
						break;
					case gamemap::HILLS:
						r = 50;
						g = 150;
						break;
					case gamemap::MOUNTAINS:
						r = 200;
						g = 100;
						break;
					case gamemap::TOWER:
						r = 200;
						g = 200;
						b = 200;
						break;
					case gamemap::CASTLE:
						r = 100;
						g = 100;
						b = 100;
						break;
				}

				*data = ((r >> minimap_->format->Rloss)
						   << minimap_->format->Rshift) |
						((g >> minimap_->format->Gloss)
						   << minimap_->format->Gshift) |
					    ((b >> minimap_->format->Bloss)
						   << minimap_->format->Bshift);
				++data;
			}
		}
	}

	while(minimap_ != NULL && minimap_->w*2 < w && minimap_->h*2 < h) {
		SDL_Surface* const surf = minimap_;
		minimap_ = scale_surface(minimap_,minimap_->w*2,minimap_->h*2);
		SDL_FreeSurface(surf);
	}

	return minimap_;
}

void display::set_paths(const paths* paths_list)
{
	pathsList_ = paths_list;
}

void display::move_unit(const std::vector<gamemap::location>& path,
						const unit& u)
{
	const SDL_Surface* const image = getImage(u.type().image());
	
	for(int i = 0; i+1 < path.size(); ++i) {
		move_unit_between(path[i],path[i+1],image);
	}
}

void display::move_unit_between(const gamemap::location& a,
                                const gamemap::location& b,
								const SDL_Surface* image)
{
	const double side_threshhold = 40.0;

	double xsrc = static_cast<double>(a.x)*zoom_*0.75 - xpos_;
	double ysrc = static_cast<double>(a.y)*zoom_ - ypos_ +
					                    ((a.x % 2) == 1 ? zoom_/2.0 : 0.0);

	double xdst = static_cast<double>(b.x)*zoom_*0.75 - xpos_;
	double ydst = static_cast<double>(b.y)*zoom_ - ypos_ +
					                    ((b.x % 2) == 1 ? zoom_/2.0 : 0.0);

	const double nsteps = 10.0;
	const double xstep = (xdst - xsrc)/nsteps;
	const double ystep = (ydst - ysrc)/nsteps;

	for(double i = 0.0; i < nsteps; i += 1.0) {
		xsrc = static_cast<double>(a.x)*zoom_*0.75 - xpos_;
		ysrc = static_cast<double>(a.y)*zoom_ - ypos_ +
						                    ((a.x % 2) == 1 ? zoom_/2.0 : 0.0);

		xdst = static_cast<double>(b.x)*zoom_*0.75 - xpos_;
		ydst = static_cast<double>(b.y)*zoom_ - ypos_ +
						                    ((b.x % 2) == 1 ? zoom_/2.0 : 0.0);

		double xloc = xsrc + xstep*i;
		double yloc = ysrc + ystep*i;
		if(xloc < side_threshhold) {
			scroll(xloc - side_threshhold,0.0);
		}

		if(yloc < side_threshhold) {
			scroll(0.0,yloc - side_threshhold);
		}

		if(xloc + double(image->w) > this->mapx() - side_threshhold) {
			scroll(((xloc + double(image->w)) - (this->mapx() - side_threshhold)),0.0);
		}

		if(yloc + double(image->h) > this->y() - side_threshhold) {
			scroll(0.0,((yloc + double(image->h)) - (this->y() - side_threshhold)));
		}

		xsrc = static_cast<double>(a.x)*zoom_*0.75 - xpos_;
		ysrc = static_cast<double>(a.y)*zoom_ - ypos_ +
						                    ((a.x % 2) == 1 ? zoom_/2.0 : 0.0);

		xdst = static_cast<double>(b.x)*zoom_*0.75 - xpos_;
		ydst = static_cast<double>(b.y)*zoom_ - ypos_ +
						                    ((b.x % 2) == 1 ? zoom_/2.0 : 0.0);

		xloc = xsrc + xstep*i;
		yloc = ysrc + ystep*i;

		draw(false);
		draw_moving_unit((int)xloc,(int)yloc,image);
		update_display();
	}
}

void display::draw_moving_unit(int x, int y, const SDL_Surface* image)
{
	const int w = mapx();
	const int h = this->y();
	if(x > w || y > h)
		return;
	
	SDL_Surface* const screen = screen_.getSurface();
	
	const short* src = reinterpret_cast<const short*>(image->pixels);
	const int endy = (y + image->h) < h ? (y + image->h) : h;
	const int endx = (x + image->w) < w ? (x + image->w) : w;
	if(endx < x)
		return;

	const int len = endx - x;
	
	if(y < 0) {
		src -= image->w*y;
		y = 0;
		if(y >= endy)
			return;
	}

	int xoffset = 0;
	if(x < 0) {
		xoffset = -x;
		x = 0;
		if(x >= endx)
			return;
	}

	for(; y != endy; ++y, src += image->w) {
		short* const dst =
				reinterpret_cast<short*>(screen->pixels) + y*screen->w + x;
		for(int i = xoffset; i != len; ++i) {
			if(src[i] != 0)
				dst[i] = src[i];
		}
	}
}

int display::calculate_energy_bar()
{
	if(energy_bar_count_ != -1) {
		return energy_bar_count_;
	}

	const SDL_Surface* const image = getImage("unmoved-energy.bmp");
	const short* const beg = reinterpret_cast<short*>(image->pixels);
	const short* const end = beg + image->w*image->h;

	const short colour = 0xFFFF;
	return energy_bar_count_ = std::count(beg,end,colour);
}
