#include "menu.hpp"

#include <unistd.h>

namespace
{
		
void draw_rectangle(int x, int y, int w, int h, short colour,
                    SDL_Surface* target)
{
	short* top_left = reinterpret_cast<short*>(target->pixels) + target->w*y+x;
	short* top_right = top_left + w;
	short* bot_left = top_left + target->w*h;
	short* bot_right = bot_left + w;

	std::fill(top_left,top_right,colour);
	std::fill(bot_left,bot_right,colour);
	while(top_left != bot_left) {
		*top_left = colour;
		*top_right = colour;
		top_left += target->w;
		top_right += target->w;
	}
}

void draw_solid_tinted_rectangle(int x, int y, int w, int h,
                                 int r, int g, int b,
                                 double alpha, SDL_Surface* target)
{
	const SDL_PixelFormat* const fmt = target->format;
	short* p = reinterpret_cast<short*>(target->pixels) + target->w*y + x;
	while(h > 0) {
		short* beg = p;
		short* const end = p + w;
		while(beg != end) {
			const int cur_r = ((*beg&fmt->Rmask) >> fmt->Rshift) << fmt->Rloss;
			const int cur_g = ((*beg&fmt->Gmask) >> fmt->Gshift) << fmt->Gloss;
			const int cur_b = ((*beg&fmt->Bmask) >> fmt->Bshift) << fmt->Bloss;

			const int new_r = int(double(cur_r)*(1.0-alpha) + double(r)*alpha);
			const int new_g = int(double(cur_g)*(1.0-alpha) + double(g)*alpha);
			const int new_b = int(double(cur_b)*(1.0-alpha) + double(b)*alpha);

			*beg = ((new_r >> fmt->Rloss) << fmt->Rshift) |
			       ((new_g >> fmt->Gloss) << fmt->Gshift) |
			       ((new_b >> fmt->Bloss) << fmt->Bshift);

			++beg;
		}
		
		p += target->w;
		--h;
	}
}
		
class menu
{
	display* display_;
	int x_, y_;
	std::vector<std::string> items_;

	short* buffer_;
	int selected_;
	bool click_selects_;
	bool previous_button_;
	bool drawn_;

	void draw_item(int item) {
		const int y = y_ + item*9;
		const int w = width();
		const int h = 9;
		
		short* const dstptr = reinterpret_cast<short*>(
						        display_->video().getSurface()->pixels) +
		                             y*display_->x() + x_;
		const short* src = buffer_ + item*9*w;

		short* dst = dstptr;

		for(int i = 0; i != h; ++i, src += w, dst += display_->x()) {
			std::copy(src,src+w,dst);
		}

		draw_solid_tinted_rectangle(x_,y,w,h,item == selected_ ? 150 : 0,0,0,
		                            0.7,display_->video().getSurface());
		display_->video().drawText(x_,y,0xFFFF,0xFFFF,items_[item].c_str());
	}

	void draw() {
		drawn_ = true;
		
		for(int i = 0; i != items_.size(); ++i)
			draw_item(i);

		display_->video().update(x_,y_,width(),height());
	}

	int hit(int x, int y) const {
		if(x > x_ && x < x_ + width() && y > y_ && y < y_ + height()){
			const int item = (y - y_)/9;
			if(item >= 0 && item < items_.size()) {
				return item;
			}
		}

		return -1;
	}
	
public:
	menu(display& disp, const std::vector<std::string>& items,
	     bool click_selects=false)
	        : display_(&disp), x_(0), y_(0), items_(items), buffer_(NULL),
	          selected_(-1), click_selects_(click_selects),
	          previous_button_(true), drawn_(false)
	{}

	~menu() {
		delete [] buffer_;
	}

	int height() const {
		return items_.size()*9;
	}

	int width() const {
		int m = 0;
		for(std::vector<std::string>::const_iterator i = items_.begin();
		    i != items_.end(); ++i) {
			if(i->size() > m)
				m = i->size();
		}

		return m*8;
	}

	int selection() const { return selected_; }

	void set_loc(int x, int y) {
		x_ = x;
		y_ = y;

		const int w = width();
		buffer_ = new short[w*height()];
		const short* src = reinterpret_cast<short*>(
						         display_->video().getSurface()->pixels) +
		                         y_*display_->x() + x_;
		
		for(int i = 0; i != height(); ++i, src += display_->x()) {
			short* const dst = buffer_ + w*i;
			const short* const src_end = src + w;
			std::copy(src,src_end,dst);
		}
	}

	int process(int x, int y, bool button) {
		const int starting_selected = selected_;

		const int hit_item = hit(x,y);
		
		if(click_selects_) {
			selected_ = hit_item;
			if(button && !previous_button_)
				return selected_;
			else {
				if(!drawn_ || selected_ != starting_selected)
					draw();
				previous_button_ = button;
				return -1;
			}
		}
		
		if(button && hit_item != -1){
			selected_ = hit_item;
		}

		if(selected_ == -1)
			selected_ = 0;

		if(selected_ != starting_selected)
			draw();

		return -1;
	}
};
		
class button
{
	display* display_;
	SDL_Surface* normal_, *mouseover_, *pressed_;
	int x_, y_;

	bool button_;

	enum STATE { UNINIT, NORMAL, ACTIVE, PRESSED };
	STATE state_;

	void draw() {
		SDL_Surface* image = normal_;
		switch(state_) {
			case ACTIVE: image = mouseover_;
						 break;
			case PRESSED: image = pressed_;
						  break;
		}

		display_->blit_surface(x_,y_,image);
		display_->video().update(x_,y_,width(),height());
	}

	bool hit(int x, int y) const {
		if(x > x_ && x < x_ + normal_->w &&
		   y > y_ && y < y_ + normal_->h) {
			x -= x_;
			y -= y_;
			int row_width = normal_->w;
			if((row_width%2) == 1)
				++row_width;

			if(*(reinterpret_cast<short*>(normal_->pixels)+y*row_width+x) != 0)
				return true;
		}

		return false;
	}
	
public:
	struct error {};
	
	button(display& disp, const std::string& button_name) : display_(&disp),
	                          x_(0), y_(0), button_(true), state_(UNINIT)
	{
		normal_ = disp.getImage(button_name + "-button.bmp",display::UNSCALED);
		mouseover_ = disp.getImage(button_name + "-button-active.bmp",
		                                                    display::UNSCALED);
		pressed_ = disp.getImage(button_name + "-button-pressed.bmp",
		                                                    display::UNSCALED);

		if(normal_ == NULL)
			throw error();

		if(mouseover_ == NULL)
			mouseover_ = normal_;

		if(pressed_ == NULL)
			pressed_ = normal_;
	}

	void set_x(int val) { x_ = val; }
	void set_y(int val) { y_ = val; }

	int width() const {
		return normal_->w;
	}

	int height() const {
		return normal_->h;
	}

	bool process(int mousex, int mousey, bool button) {
		enum MOUSE_STATE { UNCHANGED, UP, DOWN };
		MOUSE_STATE mouse_state = UNCHANGED;
		if(button && !button_)
			mouse_state = DOWN;
		else if(!button && button_)
			mouse_state = UP;

		button_ = button;
		
		const STATE start_state = state_;
		
		switch(state_) {
		case UNINIT:
			state_ = NORMAL;
			break;
		case NORMAL:
			if(hit(mousex,mousey))
				state_ = ACTIVE;
			break;
		case ACTIVE:
			if(mouse_state == DOWN && hit(mousex,mousey))
				state_ = PRESSED;
			else if(!hit(mousex,mousey))
				state_ = NORMAL;
			break;
		case PRESSED:
			if(mouse_state == UP) {
				if(hit(mousex,mousey))
					return true;
				else
					state_ = NORMAL;
			}
		}

		if(state_ != start_state)
			draw();

		return false;
	}
}; //end class button
} //end anon namespace

namespace gui
{

int show_dialog(display& disp, SDL_Surface* image,
                const std::string& caption, const std::string& message,
                DIALOG_TYPE type, const std::vector<std::string>& menu_items)
{
	CVideo& screen = disp.video();
	SDL_Surface* const scr = screen.getSurface();

	menu menu_(disp,menu_items,type == OK_ONLY || type == MESSAGE);
	
	const int border_size = 6;
	const short text_colour = 0xFFFF;
	const short border_colour = 0xF000;
	const int char_width = 8;
	const int char_height = 9;
	int nlines = 1;
	int longest_line = 0;
	int cur_line = 0;
	for(std::string::const_iterator message_it = message.begin();
	    message_it != message.end(); ++message_it) {
		if(*message_it == '\n') {
			if(cur_line > longest_line)
				longest_line = cur_line;

			cur_line = 0;

			++nlines;
		} else {
			++cur_line;
		}
	}

	const char* const * button_list = NULL;
	std::vector<button> buttons;
	switch(type) {
		case MESSAGE:
			break;
		case OK_ONLY: {
			const char* thebuttons[] = { "ok", NULL };
			button_list = thebuttons;
			break;
		}

		case YES_NO: {
			const char* thebuttons[] = { "yes", "no", NULL };
			button_list = thebuttons;
			break;
		}

		case OK_CANCEL: {
			const char* thebuttons[] = { "ok", "cancel", NULL };
			button_list = thebuttons;
			break;
		}
	}

	const int button_height_padding = 10;
	int button_heights = 0;
	int button_widths = 0;
	if(button_list != NULL) {
		try {
			while(*button_list != NULL) {
				buttons.push_back(button(disp,*button_list));

				if(buttons.back().height() > button_heights)
					button_heights = buttons.back().height();

				button_widths += buttons.back().width();

				++button_list;
			}

		} catch(button::error&) {
			//std::cerr << "error initializing button!\n";
		}
	}

	if(button_heights > 0) {
		button_heights += button_height_padding;
	}
	
	if(cur_line > longest_line)
		longest_line = cur_line;

	const int left_padding = 10;
	const int right_padding = 10;
	const int image_h_padding = image != NULL ? 10 : 0;
	const int top_padding = 10;
	const int bottom_padding = 10;
	const int padding_width = left_padding + right_padding + image_h_padding;
	const int padding_height = top_padding + bottom_padding;
	const int caption_width = caption.size()*char_width;
	const int image_width = image != NULL ? image->w : 0;
	const int total_image_width = caption_width > image_width ?
	                              caption_width : image_width;
	const int image_height = image != NULL ? image->h : 0;

	int text_width = longest_line*char_width;
	if(menu_.width() > text_width)
		text_width = menu_.width();

	int total_width = total_image_width + text_width +
	                  padding_width;
	if(button_widths > total_width)
		total_width = button_widths;

	const int total_height = (image_height+char_height > nlines*char_height ?
	                          image_height+char_height : nlines*char_height) +
	                         padding_height + button_heights + menu_.height();

	if(total_width > scr->w - 100 || total_height > scr->h - 100)
		return false;
	
	const int xloc = scr->w/2 - total_width/2;
	const int yloc = scr->h/2 - total_height/2;

	const int button_hpadding = total_width - button_widths;
	int button_offset = 0;
	for(int button_num = 0; button_num != buttons.size(); ++button_num) {
		const int padding_amount = button_hpadding/(buttons.size()+1);
		buttons[button_num].set_x(xloc + padding_amount*(button_num+1) +
		                          button_offset);
		buttons[button_num].set_y(yloc + total_height - button_heights);
		button_offset += buttons[button_num].width();
	}

	if(menu_.height() > 0)
		menu_.set_loc(xloc+total_image_width+left_padding+image_h_padding,
		              yloc+top_padding+nlines*char_height);
	
	draw_solid_tinted_rectangle(xloc-border_size,yloc-border_size,
	                            total_width+border_size,total_height+border_size,
	                            0,0,0,0.6,scr);
	draw_solid_tinted_rectangle(xloc,yloc,
	                            total_width+border_size,total_height+border_size,
								0,0,0,0.6,scr);

	draw_rectangle(xloc-border_size,yloc-border_size,
	               total_width+border_size,total_height+border_size,
	               border_colour,scr);
	draw_rectangle(xloc,yloc,total_width+border_size,total_height+border_size,
	               border_colour,scr);

	if(image != NULL) {
		const int x = xloc + left_padding;
		const int y = yloc + top_padding;

		disp.blit_surface(x,y,image);

		screen.drawText(xloc+left_padding, yloc+top_padding+image->h,
		                text_colour, text_colour, caption.c_str());
	}

	screen.drawText(xloc+total_image_width+left_padding+image_h_padding,
	                yloc+top_padding,text_colour,text_colour,message.c_str());
	
	screen.update(0,0,scr->w,scr->h);
	
	CKey key;

	bool left_button = true, right_button = true, key_down = true;
	
	for(;;) {
		int mousex, mousey;
		const int mouse_flags = SDL_GetMouseState(&mousex,&mousey);

		const bool new_right_button = mouse_flags&SDL_BUTTON_RMASK;
		const bool new_left_button = mouse_flags&SDL_BUTTON_LMASK;
		const bool new_key_down = key[KEY_SPACE] || key[KEY_ENTER] ||
		                          key[KEY_ESCAPE];

		if(menu_.height() > 0) {
			const int res = menu_.process(mousex,mousey,new_left_button);
			if(res != -1)
				return res;
		}


		if(buttons.empty() && (new_left_button && !left_button ||
		                       new_right_button && !right_button) ||
		   buttons.size() < 2 && new_key_down && !key_down)
			break;

		left_button = new_left_button;
		right_button = new_right_button;
		key_down = new_key_down;

		for(std::vector<button>::iterator button_it = buttons.begin();
		    button_it != buttons.end(); ++button_it) {
			if(button_it->process(mousex,mousey,left_button)) {
				//if the menu is not used, then return the index of the
				//button pressed, otherwise return the index of the menu
				//item selected if the last button is not pressed, and
				//cancel (-1) otherwise
				if(menu_.height() == 0)
					return button_it - buttons.begin();
				else if(button_it - buttons.begin() != buttons.size()-1)
					return menu_.selection();
				else
					return -1;
			}
		}

		SDL_PumpEvents();
	}
	
	return -1;
}

} //end namespace gui
